"""
A :class:`.GraphCollection` is a set of graphs generated from a
:class:`.Corpus` or model.
"""

import networkx as nx
from collections import defaultdict
import warnings
from tethne import networks
from tethne.utilities import _iterable

import sys
PYTHON_3 = sys.version_info[0] == 3
if PYTHON_3:
    unicode = str
    xrange = range
    str = bytes


class GraphCollection(dict):
    """
    A :class:`.GraphCollection` is an indexed set of
    :class:`.networkx.Graph`\s.

    When you add a :class:`networx.Graph`\, the nodes are indexed and
    relabeled.

    .. code-block:: python

       >>> from tethne import GraphCollection
       >>> import networkx as nx

       >>> G = GraphCollection()
       >>> g = nx.Graph()
       >>> g.add_node('A', yes='no')
       >>> g.add_edge('A', 'B', c='d')

       >>> G['graph1'] = g    # You can also use G.add('graph1', g)

       >>> G.graph1.nodes(data=True)
       [(0, {}), (1, {'yes': 'no'})]

       >>> G.node_index, G.node_lookup
       ({0: 'B', 1: 'A', -1: None}, {'A': 1, None: -1, 'B': 0})

    To build a :class:`.GraphCollection` from a :class:`.Corpus`, pass it and
    a method to the constructor, or use :meth:`.GraphCollection.build`\.

    .. code-block:: python

       >>> corpus = read(datapath)
       >>> G = GraphCollection(corpus, coauthors)

       >>> G.build(corpus, authors)

    """

    def __init__(self, corpus=None, method=None, slice_kwargs={},
                 method_kwargs={}, directed=False):
        """

        Parameters
        ----------
        corpus : :class:`.Corpus`
        method : str or func
            If str, looks for ``method`` in the ``tethne`` namespace.
        slice_kwargs : dict
            Keyword arguments to pass to ``corpus``' ``slice`` method.
        method_kwargs : dict
            Keyword arguments to pass to ``method`` along with ``corpus``.
        directed : bool
            If True, graphs will be treated as directed during indexing.
        """
        self.directed = directed
        if directed:
            self.master_graph = nx.MultiDiGraph()
        else:
            self.master_graph = nx.MultiGraph()
        self.node_index = {-1: None}
        self.node_lookup = {None: -1}
        self.graphs_containing = defaultdict(list)

        if corpus and method:
            self.build(corpus, method, slice_kwargs, method_kwargs)

    def __setitem__(self, name, graph):
        self.add(name, graph)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        try:    # hasttr() causes endless recursion in Python 3.x
            return object.__getattr__(self, name)
        except AttributeError:
            pass
        raise AttributeError('GraphCollection has no such attribute or graph.')

    def build(self, corpus, method, slice_kwargs={}, method_kwargs={}):
        """
        Generate a set of :class:`networkx.Graph`\s using ``method`` on the
        slices in ``corpus``\.

        Parameters
        ----------
        corpus : :class:`.Corpus`
        method : str or func
            If str, looks for ``method`` in the ``tethne`` namespace.
        slice_kwargs : dict
            Keyword arguments to pass to ``corpus``' ``slice`` method.
        method_kwargs : dict
            Keyword arguments to pass to ``method`` along with ``corpus``.
        """
        if not hasattr(method, '__call__'):
            if not hasattr(networks, method):
                raise NameError('No such method')
            method = getattr(networks, method)
        for key, subcorpus in corpus.slice(**slice_kwargs):
            graph = method(subcorpus, **method_kwargs)
            self.add(key, graph)

    def add(self, name, graph):
        """
        Index and add a :class:`networkx.Graph` to the
        :class:`.GraphCollection`.

        Parameters
        ----------
        name : hashable
            Unique name used to identify the `graph`.
        graph : networkx.Graph

        Raises
        ------
        ValueError
            If `name` has already been used in this :class:`.GraphCollection`\.
        """
        if name in self:
            raise ValueError("{0} exists in this GraphCollection".format(name))
        elif hasattr(self, unicode(name)):
            raise ValueError("Name conflicts with an existing attribute")

        indexed_graph = self.index(name, graph)

        # Add all edges to the `master_graph`.
        for s, t, attrs in indexed_graph.edges(data=True):
            attrs.update({'graph': name})
            self.master_graph.add_edge(s, t, **attrs)

        # Add all node attributes to the `master_graph`.
        for n, attrs in indexed_graph.nodes(data=True):
            for k,v in attrs.items():
                if k not in self.master_graph.node[n]:
                    self.master_graph.node[n][k] = {}
                self.master_graph.node[n][k][name] = v

        dict.__setitem__(self, name, indexed_graph)


    def index(self, name, graph):
        """
        Index any new nodes in `graph`, and relabel the nodes in `graph` using
        the index.

        Parameters
        ----------
        name : hashable
            Unique name used to identify the `graph`.
        graph : networkx.Graph

        Returns
        -------
        indexed_graph : networkx.Graph
        """
        nodes = graph.nodes()

        # Index new nodes.
        new_nodes = list(set(nodes) - set(self.node_index.values()))
        start = max(len(self.node_index) - 1, max(self.node_index.keys()))
        for i in xrange(start, start + len(new_nodes)):
            n = new_nodes.pop()
            self.node_index[i], self.node_lookup[n] = n, i
            self.graphs_containing[n].append(name)

        # Relabel nodes in `graph`.
        new_labels = {n: self.node_lookup[n] for n in nodes}
        indexed_graph = nx.relabel.relabel_nodes(graph, new_labels, copy=True)

        return indexed_graph

    def nodes(self, data=False, native=True):
        """
        Returns a list of all nodes in the :class:`.GraphCollection`\.

        Parameters
        ----------
        data : bool
            (default: False) If True, returns a list of 2-tuples containing
            node labels and attributes.

        Returns
        -------
        nodes : list
        """

        nodes = self.master_graph.nodes(data=data)
        if native:
            if data:
                nodes = [(self.node_index[n], attrs) for n, attrs in nodes]
            else:
                nodes = [self.node_index[n] for n in nodes]
        return nodes

    def edges(self, data=False, native=True):
        """
        Returns a list of all edges in the :class:`.GraphCollection`\.

        Parameters
        ----------
        data : bool
            (default: False) If True, returns a list of 3-tuples containing
            source and target node labels, and attributes.

        Returns
        -------
        edges : list
        """
        edges = self.master_graph.edges(data=data)
        if native:
            if data:
                edges = [(self.node_index[s], self.node_index[t], attrs)
                         for s, t, attrs in edges]
            else:
                edges = [(self.node_index[s], self.node_index[t])
                         for s, t in edges]
        return edges

    def order(self, piecewise=False):
        """
        Returns the total number of nodes in the :class:`.GraphCollection`\.
        """
        if piecewise:
            return {k: v.order() for k, v in self.items()}
        return self.master_graph.order()

    def node_distribution(self):
        warnings.warn("node_distribution will be removed in v0.8. Use" +
                      " order(piecewise=True) instead.", DeprecationWarning)
        return self.order(piecewise=True)

    def size(self, piecewise=False):
        """
        Returns the total number of edges in the :class:`.GraphCollection`\.
        """
        if piecewise:
            return {k: v.size() for k, v in self.items()}
        return self.master_graph.size()

    def edge_distribution(self):
        warnings.warn("edge_distribution will be removed in v0.8. Use" +
                      " size(piecewise=True) instead.", DeprecationWarning)
        return self.order(piecewise=True)

    def collapse(self, weight_attr='weight'):
        """
        Returns a :class:`networkx.Graph` or :class:`networkx.DiGraph` in which
        the edges between each pair of nodes are collapsed into a single
        weighted edge.
        """

        if self.directed:
            graph = nx.DiGraph()
        else:
            graph = nx.Graph()

        # Transfer all nodes and attributes.
        for n, attrs in self.master_graph.nodes(data=True):
            graph.add_node(n, attrs)

        for s, t, attrs in self.master_graph.edges(data=True):
            if not graph.has_edge(s, t):
                graph.add_edge(s, t)

            if 'weight' not in graph[s][t]:
                graph[s][t]['weight'] = 0.
            if weight_attr in attrs:
                graph[s][t]['weight'] += attrs[weight_attr]
            else:
                graph[s][t]['weight'] += 1.

            gname = attrs['graph']
            for k, v in attrs.items():
                if k in [weight_attr, 'graph']:
                    continue
                if k not in graph[s][t]:
                    graph[s][t][k] = {}
                graph[s][t][k][gname] = v
        return graph

    def analyze(self, method_name, mapper=map, invert=False, **kwargs):
        """
        Apply a method from NetworkX to each of the graphs in the
        :class:`.GraphCollection`\.

        Example
        -------
        .. code-block:: python

           >>> G.analyze('betweenness_centrality')
           {'test': {0: 1.0, 1: 0.0, 2: 0.0},
            'test2': {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0}}
           >>> G.analyze('betweenness_centrality', results_by='node')
           {0: {'test': 1.0, 'test2': 0.0},
            1: {'test': 0.0, 'test2': 0.0},
            2: {'test': 0.0, 'test2': 0.0},
            3: {'test2': 0.0}}

        Parameters
        ----------
        method : str or iterable
            Must be the name of a method accessible directly from the
            `networkx` namespace. If an iterable, should be the complete
            dot-path to the method, e.g. ``nx.connected.is_connected`` would be
            written as ``['connected', 'is_connected']``.
        mapper : func
            A mapping function. Be default uses Python's builtin ``map``
            function. MUST return results in order.
        results_by : str
            (default: 'graph'). By default, the top-level key in the results
            are graph names. If results_by='node', node labels are used as
            top-level keys.
        kwargs : kwargs
            Any additional kwargs are passed to the NetworkX method.

        Returns
        -------
        dict

        """

        # Find the analysis method, if possible.
        if hasattr(method_name, '__iter__'):
            mpath = method_name
            if type(mpath) in [str, unicode]:
                mpath = [mpath]

            root = nx
            while len(mpath) > 0:
                head = getattr(root, mpath.pop(0))
            if hasattr(head, '__call__'):
                method = head
            else:
                raise NameError("No matching method in NetworkX")

        elif hasattr(nx, method_name):
            method = getattr(nx, method_name)
        else:
            raise AttributeError('No such method in NetworkX')

        # Farm out the analysis using ``mapper``. This allows us to use
        #  multiprocessing in the future, or to add pre- or post-processing
        #  routines.
        keys, graphs = zip(*self.items())
        results = mapper(method, graphs, **kwargs)

        # Group the results by graph.
        by_graph = dict(zip(keys, results))

        # Invert results.
        inverse = defaultdict(dict)
        for gname, result in by_graph.items():
            if hasattr(result, '__iter__'):
                for n, val in result.items():
                    inverse[n].update({gname: val})

        if type(list(by_graph.values())[0]) is dict:
            # Look for a result set that we can inspect.
            i = 0
            while True:
                if len(by_graph.values()[i]) > 0:
                    inspect = by_graph.values()[i]
                    break
                i += 1

            if type(list(inspect.keys())[0]) is tuple:
                # Results correspond to edges.
                by_edge = dict(inverse)

                # Set edge attributes in each graph.
                for graph, attrs in by_graph.items():
                    nx.set_edge_attributes(self[graph], method_name, attrs)

                # Set edge attributes in the master graph.
                for (s, t), v in by_edge.items():
                    for i, attrs in self.master_graph.edge[s][t].items():
                        val = v[attrs['graph']]
                        self.master_graph.edge[s][t][i][method_name] = val

                if invert:
                    return by_edge
            else:
                # Results correspond to nodes.
                by_node = dict(inverse)

                # Set node attributes for each graph.
                for graph, attrs in by_graph.items():
                    nx.set_node_attributes(self[graph], method_name, attrs)

                # Store node attributes in the master graph.
                nx.set_node_attributes(self.master_graph, method_name, by_node)

                if invert:
                    return by_node
        return by_graph

    def node_history(self, node, attribute):
        """
        Returns a dictionary of attribute values for each ``networkx.Graph`` in
        the :class:`.GraphCollection` for a single node.

        Parameters
        ----------
        node : str
            The node of interest.
        attribute : str
            The attribute of interest; e.g. 'betweenness_centrality'

        Returns
        -------
        history : dict
        """
        return self.master_graph.node[node][attribute]

    def edge_history(self, source, target, attribute):
        """
        Returns a dictionary of attribute vales for each Graph in the
        :class:`.GraphCollection` for a single edge.

        Parameters
        ----------
        source : str
            Identifier for source node.
        target : str
            Identifier for target node.
        attribute : str
            The attribute of interest; e.g. 'betweenness_centrality'

        Returns
        -------
        history : dict
        """

        return {attr['graph']: attr[attribute] for i, attr
                in self.master_graph.edge[source][target].items()}

    def union(self, weight_attr='_weight'):
        """
        Returns the union of all graphs in this :class:`.GraphCollection`\.

        The number of graphs in which an edge exists between each node pair `u` and `v`
        is stored in the edge attribute given be `weight_attr` (default: `_weight`).

        Parameters
        ----------
        weight_attr : str
            (default: '_weight') Name of the edge attribute used to store the number of
            graphs in which an edge exists between node pairs.

        Returns
        -------
        graph : :class:`networkx.Graph`
        """

        if type(self.master_graph) is nx.MultiDiGraph:
            graph = nx.DiGraph()
        else:
            graph = nx.Graph()

        edge_attrs = defaultdict(list)

        for u, v, a in self.master_graph.edges(data=True):
            if not graph.has_edge(u, v):
                graph.add_edge(u, v)
                graph[u][v]['graphs'] = []
                graph[u][v][weight_attr] = 0.

            for key, value in a.items():
                if key not in graph[u][v]:
                    graph[u][v][key] = []
                graph[u][v][key].append(value)
            graph[u][v]['graphs'].append(a['graph'])
            graph[u][v][weight_attr] += 1.

        for u, a in self.master_graph.nodes(data=True):
            for key, value in a.items():
                graph.node[u][key] = value

        return graph



"""
Methods for building networks from bibliographic data.

Each network relies on certain meta data in the :class:`.Paper` associated with
each document. Often we wish to construct a network with nodes representing
these documents and edges representing relationships between those documents,
but this is not always the case.

Where it is the case, it is recommended but not required that nodes are
represented by an identifier from {ayjid, wosid, pmid, doi}. Each has certain
benefits. If the documents to be networked come from a single database source
such as the Web of Science, wosid is most appropriate. If not, using doi
will result in a more accurate, but also more sparse network; while ayjid
will result in a less accurate, but more complete network.

Any type of meta data from the :class:`.Paper` may be used as an identifier,
however.

We use "head" and "tail" nomenclature to refer to the members of a directed
edge (x,y), x -> y, xy, etc. by calling x the "tail" and y the "head".

Modules
```````

.. autosummary::

   authors
   features
   papers
   topics

"""

from tethne.networks.authors import author_papers, coauthors
from tethne.networks.papers import author_coupling, bibliographic_coupling, \
                                   cocitation, direct_citation
from tethne.networks.topics import topic_coupling, cotopics, terms



"""
Methods for generating networks in which papers are vertices.

.. autosummary::
   :nosignatures:

   author_coupling
   bibliographic_coupling
   cocitation
   direct_citation
   topic_coupling

"""

from tethne.networks.base import multipartite, coupling, cooccurrence


def direct_citation(corpus, min_weight=1):
    """
    A directed paper-citation network.

    Direct-citation graphs are `directed acyclic graphs`__ in which vertices
    are papers, and each (directed) edge represents a citation of the target
    paper by the source paper. The :func:`.networks.papers.direct_citation`
    method generates both a global citation graph, which includes all cited and
    citing papers, and an internal citation graph that describes only citations
    among papers in the original dataset.
    """

    return multipartite(corpus, ['citations'], min_weight=min_weight)


def bibliographic_coupling(corpus, min_weight=1):
    """
    Generate a bibliographic coupling network.

    Two papers are **bibliographically coupled** when they both cite the same,
    third, paper.
    """
    return coupling(corpus, 'citations', min_weight=min_weight)


def cocitation(corpus, min_weight=1, edge_attrs=['ayjid', 'date']):
    """
    Generate a cocitation network.

    A **cocitation network** is a network in which vertices are papers, and
    edges indicate that two papers were cited by the same third paper.
    `CiteSpace
    <http://cluster.cis.drexel.edu/~cchen/citespace/doc/jasist2006.pdf>`_
    is a popular desktop application for co-citation analysis, and you can read
    about the theory behind it
    `here <http://cluster.cis.drexel.edu/~cchen/citespace/>`_.
    """
    return cooccurrence(corpus, 'citations', min_weight=min_weight,
                        edge_attrs=edge_attrs)


def author_coupling(corpus, min_weight=1):
    return coupling(corpus, 'authors', min_weight=min_weight)



"""
Methods for generating networks in which authors are vertices.

.. autosummary::
   :nosignatures:

   author_papers
   coauthors

"""

from tethne.networks.base import cooccurrence, multipartite


def author_papers(corpus, min_weight=1):
    """
    A bi-partite graph containing :class:`.Paper`\s and their authors.
    """
    return multipartite(corpus, ['authors'], min_weight=min_weight)


def coauthors(corpus, min_weight=1, edge_attrs=['ayjid', 'date']):
    """
    A graph describing joint authorship in ``corpus``.
    """
    return cooccurrence(corpus, 'authors', min_weight=min_weight,
                        edge_attrs=edge_attrs)



from setuptools import setup, Extension

DISTNAME = 'tethne'
AUTHOR = 'E. Peirson, Digital Innovation Group @ ASU'
MAINTAINER = 'Erick Peirson'
MAINTAINER_EMAIL = 'erick [dot] peirson [at] asu [dot] edu'
DESCRIPTION = ('Bibliographic network and corpus analysis for historians')
LICENSE = 'GNU GPL 3'
URL = 'http://diging.github.io/tethne/'
VERSION = '0.7.17'

PACKAGES = ['tethne',
            'tethne.analyze',
            'tethne.classes',
            'tethne.model',
            'tethne.model.corpus',
            'tethne.networks',
            'tethne.readers',
            'tethne.writers',
            'tethne.plot']

import sys
if sys.version_info[0] == 2:
    pdfminer = 'pdfminer==20110515'
elif sys.version_info[0] == 3:
    pdfminer = "pdfminer3k"


setup(
    name=DISTNAME,
    author=AUTHOR,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    description=DESCRIPTION,
    license=LICENSE,
    url=URL,
    version=VERSION,
    packages = PACKAGES,
    include_package_data=True,
    install_requires=[
        "networkx >= 1.8.1",
        "Unidecode >= 0.04.16",
        "iso8601",
        "rdflib",
        "chardet",
        "html5lib",
        "isodate",
        "pdfdocument",
        pdfminer,
        "python-magic==0.4.6",
        "slate",
        "Unidecode==0.4.17",
        "nltk",
    ],
)



"""
Write :class:`.GraphCollection` to a structured data format.

.. autosummary::

   to_dxgmml

"""

import sys
PYTHON_3 = sys.version_info[0] == 3
if PYTHON_3:
    unicode = str

import networkx as nx
import pickle as pk


def to_dxgmml(graphcollection, path): # [#61510094]
    """
    Writes a :class:`.GraphCollection` to
    `dynamic XGMML. <https://code.google.com/p/dynnetwork/wiki/DynamicXGMML>`_.

    Dynamic XGMML is a schema for describing dynamic networks in Cytoscape 3.0.
    This method assumes that `Graph` indices are orderable points in time
    (e.g. years). The "start" and "end" of each node and edge are determined by
    periods of consecutive appearance in the :class:`.GraphCollection` . Node
    and edge attributes are defined for each `Graph`. in the
    :class:`.GraphCollection`.

    For example, to build and visualize an evolving co-citation network:

    .. code-block:: python

       >>> # Load some data.
       >>> import tethne.readers as rd
       >>> papers = rd.wos.read(datapath)

       >>> # Build a Corpus, and slice it temporally using a
       >>> #  4-year sliding time-window.
       >>> from tethne.data import Corpus, GraphCollection
       >>> D = Corpus(papers)
       >>> D.slice('date', 'time_window', window_size=4)

       >>> # Generate a GraphCollection of co-citation graphs.
       >>> from tethne.builders import paperCollectionBuilder
       >>> builder = paperCollectionBuilder(D)
       >>> C = builder.build('date', 'cocitation', threshold=2)

       >>> # Write the GraphCollection as a dynamic network.
       >>> import tethne.writers as wr
       >>> wr.collection.to_dxgmml(C, "/path/to/network.xgmml")

    Parameters
    ----------
    graphcollection : :class:`.GraphCollection`
        The :class:`.GraphCollection` to be written to XGMML.
    path : str
        Path to file to be written. Will be created/overwritten.

    Raises
    ------
    AttributeError
        C must be a tethne.classes.GraphCollection.

    Notes
    -----
    Period start and end dates in this method are inclusive, whereas XGMML end
    dates are exclusive. Hence +1 is added to all end dates when writing XGMML.
    """

    # TODO: make sure C is a GraphCollection.

    nodes = {}
    for n in graphcollection.nodes():
        nodes[n] = { 'periods' : [] }   # Each period will be a dict with
                                        #  'start' and 'end' values.
    edges = {}
    for e in graphcollection.edges():
        edges[e] = { 'periods' : [] }

    # Build node list.
    current = []
    for k in sorted(graphcollection.keys()):
        graph = _strip_list_attributes(graphcollection[k])
        preceding = [ c for c in current ]
        current = []
        for n in graph.nodes(data=True):
            n_ = graphcollection.node_index[n[0]]
            if n_ not in preceding:   # Looking for gaps in presence of node.
                nodes[n_]['periods'].append({'start': k, 'end': k})
            else:
                if k > nodes[n_]['periods'][-1]['end']:
                    nodes[n_]['periods'][-1]['end'] = k
            current.append(n_)

            nodes[n_][k] = {}
            for attr, value in n[1].items():
                if type(value) is str:
                    value = value.replace("&", "&amp;").replace('"', '')
                nodes[n_][k][attr] = value

    # Build edge list.
    current = []
    for k in sorted(graphcollection.keys()):
        graph = _strip_list_attributes(graphcollection[k])
        preceding = [ c for c in current ]
        current = []
        for e in graph.edges(data=True):
            e_key = (graphcollection.node_index[e[0]],
                     graphcollection.node_index[e[1]])
            if e_key not in preceding:   # Looking for gaps in presence of edge.
                try:
                    edges[e_key]['periods'].append( { 'start': k, 'end': k } )
                except KeyError:
                    e_key = (graphcollection.node_index[e[1]],
                             graphcollection.node_index[e[0]])
                    edges[e_key]['periods'].append( { 'start': k, 'end': k } )
            else:
                if k > edges[e_key]['periods'][-1]['end']:
                    edges[e_key]['periods'][-1]['end'] = k
            current.append(e_key)

            edges[e_key][k] = {}
            for attr, value in e[2].items():
                if type(value) is str:
                    value = value.replace("&", "&amp;").replace('"', '')
                edges[e_key][k][attr] = value

    # Write graph to XGMML.
    xst = '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n'

    sg = '<graph>\n'
    eg = '</graph>'

    nst = '\t<node label="{0}" id="{0}" start="{1}" end="{2}">\n'
    ast = '\t\t<att name="{0}" type="{1}" value="{2}" start="{3}" end="{4}"/>\n'
    enn = '\t</node>\n'

    est = '\t<edge source="{0}" target="{1}" start="{2}" end="{3}">\n'
    eas = '\t\t<att name="{0}" type="{1}" value="{2}" start="{3}" end="{4}"/>\n'
    ene = '\t</edge>\n'

    with open(path, "w") as f:
        f.write(xst)    # xml element.
        f.write(sg)     # Graph element.
        for n in nodes.keys():
            for period in nodes[n]['periods']:
                label = unicode(n).replace("&", "&amp;").replace('"', '')

                # Node element.
                f.write(nst.format(label, period['start'], period['end']+1))

                for i in sorted(nodes[n].keys()):
                    if period['start'] <= i <= period['end']:
                        for attr, value in nodes[n][i].items():
                            # Type names are slightly different in XGMML.
                            dtype = _safe_type(value)
                            attr = unicode(attr).replace("&", "&amp;")

                            # Node attribute element.
                            f.write(ast.format(attr, dtype, value, i, i+1))

                f.write(enn)    # End node element.

        for e in edges.keys():
            for period in edges[e]['periods']:
                src = unicode(e[0]).replace("&", "&amp;").replace('"', '')
                tgt = unicode(e[1]).replace("&", "&amp;").replace('"', '')
                start = period['start']
                end = period['end'] + 1

                # Edge element.
                f.write(est.format(src, tgt, start, end))

                for i in sorted(edges[e].keys()):
                    if period['start'] <= i <= period['end']:
                        for attr, value in edges[e][i].items():
                            # Type names are slightly different in XGMML.
                            dtype = _safe_type(value)

                            # Edge attribute element.
                            f.write(eas.format(attr, dtype, value, i, i+1)
                                       .replace("&", "&amp;"))

                f.write(ene)    # End edge element.
        f.write(eg) # End graph element.

def _strip_list_attributes(graph_):
    """Converts lists attributes to strings for all nodes and edges in G."""
    for n_ in graph_.nodes(data=True):
        for k,v in n_[1].items():
            if type(v) is list:
                graph_.node[n_[0]][k] = unicode(v)
    for e_ in graph_.edges(data=True):
        for k,v in e_[2].items():
            if type(v) is list:
                graph_.edge[e_[0]][e_[1]][k] = unicode(v)

    return graph_

def _safe_type(value):
    """Converts Python type names to XGMML-safe type names."""

    if type(value) is str: dtype = 'string'
    if type(value) is unicode: dtype = 'string'
    if type(value) is int: dtype = 'integer'
    if type(value) is float: dtype = 'real'

    return dtype



"""
This module provides :class:`.Corpus`\.
"""

from collections import Counter
from itertools import chain
import hashlib
from math import log

from tethne.classes.feature import FeatureSet, Feature, \
                                   StructuredFeatureSet, StructuredFeature
from tethne.utilities import _iterable, argsort

import sys
PYTHON_3 = sys.version_info[0] == 3
if PYTHON_3:
    unicode = str


def _tfidf(f, c, C, DC, N):
    tf = float(c)
    idf = log(float(N)/float(DC))
    return tf*idf


def _filter(s, C, DC):
    if C > 3 and DC > 1 and len(s) > 3:
        return True
    return False


class Corpus(object):
    """
    A :class:`.Corpus` represents a collection of :class:`.Paper` instances.

    .. autosummary::
       :toctree:
       :nosignatures:

       distribution
       feature_distribution
       features
       index
       index_by
       index_feature
       indexed_papers
       indices
       papers
       select
       slice
       structuredfeatures
       subcorpus
       top_features

    :class:`.Corpus` objects are generated by the bibliographic readers in the
    :mod:`tethne.readers` module.

    .. code-block:: python

       >>> from tethne.readers.wos import read
       >>> read('/path/to/data')
       <tethne.classes.corpus.Corpus object at 0x10278ea10>

    You can also build a :class:`.Corpus` from a list of :class:`.Paper`\s.

    .. code-block:: python

       >>> papers = however_you_generate_papers()   # <- list of Papers.
       >>> corpus = Corpus(papers)

    All of the :class:`.Paper`\s in the :class:`.Corpus` will be indexed. You
    can control which field is used for indexing by passing the ``index_by``
    keyword argument to one of the ``read`` methods or to the :class:`.Corpus`
    constructor.

    .. code-block:: python

       >>> corpus = Corpus(papers, index_by='doi')
       >>> corpus.indexed_papers.keys()
       ['doi/123', 'doi/456', ..., 'doi/789']

    The WoS ``read`` method uses the ``wosid`` field by default, and the DfR
    ``read`` method uses ``doi``. The Zotero ``read`` method tries to use
    whatever it can find. If the selected ``index_by`` field is not set or not
    available, a unique key will be generated using the title and author names.

    By default, :class:`.Corpus` will also index the ``authors`` and
    ``citations`` fields. To control which fields are indexed, pass the
    ``index_fields`` argument, or call :meth:`.Corpus.index` directly.

    .. code-block:: python

       >>> corpus = Corpus(papers, index_fields=['authors', 'date'])
       >>> corpus.indices.keys()
       ['authors', 'date']

    Similarly, :class:`.Corpus` will index features. By default, ``authors``
    and ``citations`` will be indexed as features (i.e. available for
    network-building methods). To control which fields are indexed as features,
    pass the ``index_features`` argument, or call
    :meth:`.Corpus.index_features`\.

    .. code-block:: python

       >>> corpus = Corpus(papers, index_features=['unigrams'])
       >>> corpus.features.keys()
       ['unigrams']

    There are a variety of ways to select :class:`.Paper`\s from the corpus.

    .. code-block:: python

       >>> corpus = Corpus(papers)
       >>> corpus[0]    # Integer indices yield a single Paper.
       <tethne.classes.paper.Paper object at 0x103037c10>

       >>> corpus[range(0,5)]  # A list of indices will yield a list of Papers.
       [<tethne.classes.paper.Paper object at 0x103037c10>,
        <tethne.classes.paper.Paper object at 0x10301c890>,
        ...
        <tethne.classes.paper.Paper object at 0x10302f5d0>]

       >>> corpus[('date', 1995)]  # You can select based on indexed fields.
       [<tethne.classes.paper.Paper object at 0x103037c10>,
        <tethne.classes.paper.Paper object at 0x10301c890>,
        ...
        <tethne.classes.paper.Paper object at 0x10302f5d0>]

       >>> corpus['citations', ('DOLE RJ 1952 CELL')]   # All papers with this citation!
       [<tethne.classes.paper.Paper object at 0x103037c10>,
        <tethne.classes.paper.Paper object at 0x10301c890>,
        ...
        <tethne.classes.paper.Paper object at 0x10302f5d0>]

       >>> corpus[('date', range(1993, 1995))]  # Multiple values are supported, too.
       [<tethne.classes.paper.Paper object at 0x103037c10>,
        <tethne.classes.paper.Paper object at 0x10301c890>,
        ...
        <tethne.classes.paper.Paper object at 0x10302f5d0>]

    If you prefer to retrieve a :class:`.Corpus` rather than simply a list of
    :class:`.Paper` instances (e.g. to build networks), use
    :meth:`.Corpus.subcorpus`\. ``subcorpus`` accepts selector arguments
    just like :meth:`.Corpus.__getitem__`\.

    .. code-block:: python

       >>> corpus = Corpus(papers)
       >>> subcorpus = corpus.subcorpus(('date', 1995))
       >>> subcorpus
       <tethne.classes.corpus.Corpus object at 0x10278ea10>

    """
    @property
    def papers(self):
        """
        A list of all :class:`.Paper`\s in the :class:`.Corpus`\.
        """
        return list(self.indexed_papers.values())

    index_by = None
    """
    Specifies the field in :class:`.Paper`\s that should be used as the
    primary indexing field for a :class:`.Corpus` instance.
    """

    indexed_papers = {}
    """
    The primary index for :class:`.Paper`\s in a :class:`.Corpus` instance.
    Keys are based on :attr:`.index_by`\, and values are :class:`.Paper`
    instances.
    """

    features = {}
    """
    Contains :class:`.FeatureSet`\s for a :class:`.Corpus` instance.

    New :class:`.FeatureSet`\s can be created from attributes of :class:`.Paper`
    using :meth:`.index_feature`\.
    """

    indices = {}
    """
    Contains field indices for the :class:`Paper`\s in a :class:`.Corpus`
    instance.

    The ``'citations'`` index, for example, allows us to look up all of the
    Papers that contain a particular bibliographic reference:

    .. code-block:: python

       >>> for citation, papers in corpus.indices['citations'].items()[7:10]:
       ...     print 'The following Papers cite {0} \\n\\n\\t{1} \\n'.format(citation, '\\n\\t'.join(papers))
       The following Papers cite WHITFIELD J 2006 NATURE:
            WOS:000252758800011
            WOS:000253464000004
       The following Papers cite WANG T 2006 GLOBAL CHANGE BIOL:
            WOS:000282225000019
            WOS:000281546800001
            WOS:000251903200006
            WOS:000292901400010
            WOS:000288656800015
            WOS:000318353300001
            WOS:000296710600017
            WOS:000255552100006
            WOS:000272153800012
       The following Papers cite LINKOSALO T 2009 AGR FOREST METEOROL:
            WOS:000298398700003

    Notice that the values above are not Papers themselves, but identifiers.
    These are the same identifiers used in the primary index, so we can use them
    to look up :class:`Paper`\s:

       >>> papers = corpus.indices['citations']['CARLSON SM 2004 EVOL ECOL RES']  # Who cited Carlson 2004?
       >>> print papers
       >>> for paper in papers:
       ...     print corpus.indexed_papers[paper]
       ['WOS:000311994600006', 'WOS:000304903100014', 'WOS:000248812000005']
       <tethne.classes.paper.Paper object at 0x112d1fe10>
       <tethne.classes.paper.Paper object at 0x1121e8310>
       <tethne.classes.paper.Paper object at 0x1144ad390>

    You can create new indices using :meth:`.index`.
    """

    def __init__(self, papers=None, index_by=None,
                 index_fields=['authors', 'citations'],
                 index_features=['authors', 'citations'], **kwargs):
        """
        Parameters
        ----------
        paper : list
        index_by : str
        index_fields : str or iterable of strs
        kwargs : kwargs

        """

        self.index_by = index_by
        self.slices = []
        self.indices = {}
        self.features = {}

        self.indexed_papers = {self._generate_index(paper): paper
                               for paper in papers}

        if index_features:
            for feature_name in index_features:
                self.index_feature(feature_name)

        if index_fields:
            for attr in _iterable(index_fields):
                self.index(attr)


    def __len__(self):
        return len(self.indexed_papers)

    def _generate_index(self, paper):
        """
        If the ``index_by`` field is not set or not available, generate a unique
        identifier using the :class:`.Paper`\'s title and author names.
        """
        if self.index_by is None or not hasattr(paper, self.index_by):
            if not hasattr(paper, 'hashIndex'): # Generate a new index for this paper.
                m = hashlib.md5()

                # If we dont have author name then we just use the title of the paper
                # to generate unique identifier.
                if paper.authors is None:
                    hashable = paper.title
                elif len(paper.authors) == 0:
                    hashable = paper.title
                else:
                    authors = list(zip(*paper.authors))[0]
                    hashable = u' '.join(list([paper.title] + [l + f for l, f in authors]))

                m.update(hashable.encode('utf-8'))
                setattr(paper, 'hashIndex', m.hexdigest())
            return getattr(paper, 'hashIndex')
        return getattr(paper, self.index_by)    # Identifier is already available.

    def index_feature(self, feature_name, tokenize=lambda x: x, structured=False):
        """
        Creates a new :class:`.FeatureSet` from the attribute ``feature_name``
        in each :class:`.Paper`\.

        New :class:`.FeatureSet`\s are added to :attr:`.features`\.

        Parameters
        ----------
        feature_name : str
            The name of a :class:`.Paper` attribute.

        """
        if structured:
            fclass = StructuredFeature
            fsclass = StructuredFeatureSet
        else:
            fclass = Feature
            fsclass = FeatureSet

        feats = {self._generate_index(p): fclass(tokenize(getattr(p, feature_name)))
                 for p in self.papers if hasattr(p, feature_name)}
        self.features[feature_name] = fsclass(feats)

    def index(self, attr):
        """
        Indexes the :class:`.Paper`\s in this :class:`.Corpus` instance
        by the attribute ``attr``.

        New indices are added to :attr:`.indices`\.

        Parameters
        ----------
        attr : str
            The name of a :class:`.Paper` attribute.

        """

        self.indices[attr] = {}
        for i, paper in self.indexed_papers.items():
            if hasattr(paper, attr):
                value = getattr(paper, attr)
                for v in _iterable(value):
                    if type(value) is Feature:
                        v_ = v[:-1]
                    else:
                        v_ = v

                    if hasattr(v_, '__iter__'):
                        if len(v_) == 1:
                            t = type(v_[0])
                            v_ = t(v_[0])

                    if v_ not in self.indices[attr]:
                        self.indices[attr][v_] = []
                    self.indices[attr][v_].append(i)


    def __getitem__(self, selector):
        return self.select(selector)

    def __getattr__(self, key):
        if hasattr(self, key):
            return getattr(self, key)
        elif key in self.indices:
            return self.indices[key]
        raise AttributeError("Corpus has no such attribute")

    def select(self, selector):
        """
        Retrieves a subset of :class:`.Paper`\s based on selection criteria.

        There are a variety of ways to select :class:`.Paper`\s.

        .. code-block:: python

           >>> corpus = Corpus(papers)
           >>> corpus[0]    # Integer indices yield a single Paper.
           <tethne.classes.paper.Paper object at 0x103037c10>

           >>> corpus[range(0,5)]  # A list of indices yields a list of Papers.
           [<tethne.classes.paper.Paper object at 0x103037c10>,
            <tethne.classes.paper.Paper object at 0x10301c890>,
            ...
            <tethne.classes.paper.Paper object at 0x10302f5d0>]

           >>> corpus[('date', 1995)]  # Select based on indexed fields.
           [<tethne.classes.paper.Paper object at 0x103037c10>,
            <tethne.classes.paper.Paper object at 0x10301c890>,
            ...
            <tethne.classes.paper.Paper object at 0x10302f5d0>]

           >>> corpus['citations', ('DOLE RJ 1952 CELL')]   # Citing papers!
           [<tethne.classes.paper.Paper object at 0x103037c10>,
            <tethne.classes.paper.Paper object at 0x10301c890>,
            ...
            <tethne.classes.paper.Paper object at 0x10302f5d0>]

           >>> corpus[('date', range(1993, 1995))] # Multiple values are OK.
           [<tethne.classes.paper.Paper object at 0x103037c10>,
            <tethne.classes.paper.Paper object at 0x10301c890>,
            ...
            <tethne.classes.paper.Paper object at 0x10302f5d0>]

        If you prefer to retrieve a :class:`.Corpus` rather than simply a
        list of :class:`.Paper` instances (e.g. to build networks), use
        :meth:`.Corpus.subcorpus`\.

        Parameters
        ----------
        selector : object
            See method description.

        Returns
        -------
        list
            A list of :class:`.Paper`\s.
        """

        papers = []
        if type(selector) is tuple: # Select papers by index.
            index, value = selector
            if type(value) is list:  # Set of index values.
                papers = [p for v in value for p in self[index, v]]
            else:
                if value in self.indices[index]:
                    papers = [self.indexed_papers[p] for p  # Single index value.
                              in self.indices[index][value]]
                else:
                    papers = []
        elif type(selector) is list:
            if selector[0] in self.indexed_papers:
                # Selector is a list of primary indices.
                papers = [self.indexed_papers[s] for s in selector]
            elif type(selector[0]) is int:
                papers = [self.papers[i] for i in selector]
        elif type(selector) is int:
            papers = self.papers[selector]
        elif type(selector) in [str, unicode]:
            if selector in self.indexed_papers:
                papers = self.indexed_papers[selector]
        return papers

    def slice(self, window_size=1, step_size=1, cumulative=False, count_only=False, subcorpus=True):
        """
        Returns a generator that yields ``(key, subcorpus)`` tuples for
        sequential time windows.

        Two common slicing patterns are the "sliding time-window" and the
        "time-period" patterns. Whereas time-period slicing divides the corpus
        into subcorpora by sequential non-overlapping time periods, subcorpora
        generated by time-window slicing can overlap.

        .. figure:: _static/images/bibliocoupling/timeline.timeslice.png
           :width: 400
           :align: center

           **Time-period** slicing, with a window-size of 4 years.

        .. figure:: _static/images/bibliocoupling/timeline.timewindow.png
           :width: 400
           :align: center

           **Time-window** slicing, with a window-size of 4 years and a
           step-size of 1 year.

        *Sliding time-window* -- Set ``step_size=1``, and ``window_size`` to
        the desired value.
        *Time-period* -- ``step_size`` and ``window_size`` should have the same
        value.

        The value of ``key`` is always the first year in the slice.

        Example
        -------
        .. code-block:: python

           >>> from tethne.readers.wos import read
           >>> corpus = read('/path/to/data')
           >>> for key, subcorpus in corpus.slice():
           ...     print key, len(subcorpus)
           2005, 5
           2006, 5

        Parameters
        ----------
        window_size : int
            (default: 1) Size of the time window, in years.
        step_size : int
            (default: 1) Number of years to advance window at each step.

        Returns
        -------
        generator
        """

        if 'date' not in self.indices:
            self.index('date')

        start = min(self.indices['date'].keys())
        end = max(self.indices['date'].keys())

        while start <= end - (window_size - 1):
            selector = ('date', range(start, start + window_size, 1))
            if cumulative:
                year = start + window_size
            else:
                year = start
            if count_only:
                yield year, len(self.select(selector))
            elif subcorpus:
                yield year, self.subcorpus(selector)
            else:
                yield year, self.select(selector)
            if cumulative:
                window_size += step_size
            else:
                start += step_size

    def distribution(self, **slice_kwargs):
        """
        Calculates the number of papers in each slice, as defined by
        ``slice_kwargs``.

        Example
        -------
        .. code-block:: python

           >>> corpus.distribution(step_size=1, window_size=1)
           [5, 5]

        Parameters
        ----------
        slice_kwargs : kwargs
            Keyword arguments to be passed to :method:`.Corpus.slice`\.

        Returns
        -------
        list
        """
        values = []
        keys = []

        for key, size in self.slice(count_only=True, **slice_kwargs):
            values.append(size)
            keys.append(key)
        return keys, values

    def feature_distribution(self, featureset_name, feature, mode='counts',
                             **slice_kwargs):
        """
        Calculates the distribution of a feature across slices of the corpus.

        Example
        -------
        .. code-block:: python

           >>> corpus.feature_distribution(featureset_name='citations', \
           ...                             feature='DOLE RJ 1965 CELL', \
           ...                             step_size=1, window_size=1)
           [2, 15, 25, 1]

        Parameters
        ----------
        featureset_name : str
            Name of a :class:`.FeatureSet` in the :class:`.Corpus`\.
        feature : str
            Name of the specific feature of interest. E.g. if
            ``featureset_name='citations'``, then ``feature`` could be
            something like ``'DOLE RJ 1965 CELL'``.
        mode : str
            (default: ``'counts'``) If set to ``'counts'``, values will be the
            sum of all count values for the feature in each slice. If set to
            ``'documentCounts'``, values will be the number of papers in which
            the feature occurs in each slice.
        slice_kwargs : kwargs
            Keyword arguments to be passed to :method:`.Corpus.slice`\.

        Returns
        -------
        list
        """

        values = []
        keys = []
        fset = self.features[featureset_name]

        for key, papers in self.slice(subcorpus=False, **slice_kwargs):
            allfeatures = [v for v
                           in chain(*[fset.features[self._generate_index(p)]
                                      for p in papers
                                      if self._generate_index(p)
                                      in fset.features])]

            if len(allfeatures) < 1:
                keys.append(key)
                values.append(0.)
                continue

            count = 0.
            for elem, v in allfeatures:
                if elem != feature:
                    continue
                if mode == 'counts':
                    count += v
                else:
                    count += 1.
            values.append(count)
            keys.append(key)
        return keys, values

    def top_features(self, featureset_name, topn=20, by='counts',
                     perslice=False, slice_kwargs={}):
        """
        Retrieves the top ``topn`` most numerous features in the corpus.

        Parameters
        ----------
        featureset_name : str
            Name of a :class:`.FeatureSet` in the :class:`.Corpus`\.
        topn : int
            (default: ``20``) Number of features to return.
        by : str
            (default: ``'counts'``) If ``'counts'``, uses the sum of feature
            count values to rank features. If ``'documentCounts'``, uses the
            number of papers in which features occur.
        perslice : bool
            (default: False) If True, retrieves the top ``topn`` features in
            each slice.
        slice_kwargs : kwargs
            If ``perslice=True``, these keyword arguments are passed to
            :meth:`.Corpus.slice`\.
        """

        if perslice:
            return [(k, subcorpus.features[featureset_name].top(topn, by=by))
                    for k, subcorpus in self.slice(**slice_kwargs)]
        return self.features[featureset_name].top(topn, by=by)

    def subcorpus(self, selector):
        """
        Generates a new :class:`.Corpus` using the criteria in ``selector``.

        Accepts selector arguments just like  :meth:`.Corpus.select`\.

        .. code-block:: python

           >>> corpus = Corpus(papers)
           >>> subcorpus = corpus.subcorpus(('date', 1995))
           >>> subcorpus
           <tethne.classes.corpus.Corpus object at 0x10278ea10>

        """

        subcorpus = Corpus(self[selector],
                           index_by=self.index_by,
                           index_fields=self.indices.keys(),
                           index_features=self.features.keys())

        # Transfer FeatureSets.
        for featureset_name, featureset in self.features.items():
            if featureset_name not in subcorpus:
                shared_papers = list(set(featureset.features.keys()) & \
                                     set(subcorpus.indexed_papers.keys()))
                fset = FeatureSet({k: featureset.features[k]
                                   for k in shared_papers})
                subcorpus.features[featureset_name] = fset

        return subcorpus



"""
Corpus models describe latent topics (dimensions) that explain the
distribution of features (eg words) among documents in a :class:`.Corpus`\.

Tethne presently represents two corpus models:

.. autosummary::
   :nosignatures:
   
   ldamodel.LDAModel
   dtmmodel.DTMModel
   
Most model classes are subclasses of :class:`.BaseModel`\. It is assumed that
each model describes a set of items (eg :class:`.Paper`\s or authors), a set 
of dimensions that describe those items (eg topics), and a set of features 
that comprise those dimensions (eg words).
"""

#from dtmmodel import *
#from ldamodel import *


"""
Methods for analyzing :class:`.GraphCollection`\s.

.. autosummary::
   :nosignatures:
   
   algorithm
   attachment_probability
   connected
   delta
   node_global_closeness_centrality
   
"""

import networkx
import graph
import warnings

def algorithm(G, method_name, **kwargs):
    """
    Apply a ``method`` from NetworkX to all ``networkx.Graph`` objects in the
    :class:`.GraphCollection` ``G``.
    
    For options, see the `list of algorithms
    <http://networkx.github.io/documentation/networkx-1.9/reference/algorithms.html>`_
    in the NetworkX documentation. Not all of these have been tested.

    Parameters
    ----------
    G : :class:`.GraphCollection`
        The :class:`.GraphCollection` to analyze. The specified method will be
        applied to each graph in ``G``.
    method_name : string
        Name of a method in NetworkX to execute on graph collection.
    **kwargs
        A list of keyword arguments that should correspond to the parameters
        of the specified method.

    Returns
    -------
    results : dict
        Indexed by element (node or edge) and graph index (e.g. ``date``).

    Raises
    ------
    ValueError
        If no such method exists.

    Examples
    --------

    *Betweenness centrality:* (``G`` is a :class:`.GraphCollection`\)

    .. code-block:: python

       >>> from tethne.analyze import collection
       >>> BC = collection.algorithm(G, 'betweenness_centrality')
       >>> print BC[0]
       {1999: 0.010101651117889644,
       2000: 0.0008689093723107329,
       2001: 0.010504898852426189,
       2002: 0.009338654511194512,
       2003: 0.007519105636349891}

    """
    warnings.warn("To be removed in 0.8. Use GraphCollection.analyze instead.",
                  DeprecationWarning)

    return G.analyze(method_name, **kwargs)


def connected(G, method_name, **kwargs):
    """
    Performs analysis methods from networkx.connected on each graph in the
    collection.

    Parameters
    ----------
    G : :class:`.GraphCollection`
        The :class:`.GraphCollection` to analyze. The specified method will be
        applied to each graph in ``G``.
    method : string
        Name of method in networkx.connected.
    **kwargs : kwargs
        Keyword arguments, passed directly to method.

    Returns
    -------
    results : dict
        Keys are graph indices, values are output of method for that graph.

    Raises
    ------
    ValueError
        If name is not in networkx.connected, or if no such method exists.

    """
    warnings.warn("To be removed in 0.8. Use GraphCollection.analyze instead.",
                  DeprecationWarning)

    return G.analyze(['connected', method_name], **kwargs)


def attachment_probability(G):
    """
    Calculates the observed attachment probability for each node at each
    time-step.
    
    
    Attachment probability is calculated based on the observed new edges in the
    next time-step. So if a node acquires new edges at time t, this will accrue
    to the node's attachment probability at time t-1. Thus at a given time,
    one can ask whether degree and attachment probability are related.

    Parameters
    ----------
    G : :class:`.GraphCollection`
        Must be sliced by 'date'. See :func:`.GraphCollection.slice`\.
    
    Returns
    -------
    probs : dict
        Keyed by index in G.graphs, and then by node.
    """
    warnings.warn("Removed in 0.8. Too domain-specific.")

    probs = {}
    G_ = None
    k_ = None
    for k,g in G.graphs.items():
        new_edges = {}
        if G_ is not None: 
            for n in g.nodes():
                try:
                    old_neighbors = set(G_[n].keys())
                    if len(old_neighbors) > 0:
                        new_neighbors = set(g[n].keys()) - old_neighbors
                        new_edges[n] = float(len(new_neighbors))
                    else:
                        new_edges[n] = 0.
                except KeyError:
                    pass
    
            N = sum( new_edges.values() )
            probs[k_] = { n:0. for n in G_.nodes() }
            if N > 0.:
                for n in G.nodes():
                    try:
                        probs[k_][n] = new_edges[n]/N
                    except KeyError:
                        pass

            if probs[k_] is not None:
                networkx.set_node_attributes(G.graphs[k_],
                                             'attachment_probability',
                                             probs[k_])
    
        G_ = G
        k_ = k

    # Handle last graph (no values).
    key = G.graphs.keys()[-1]
    zprobs = { n:0. for n in G.graphs[key].nodes() }
    networkx.set_node_attributes(G.graphs[key], 'attachment_probability', zprobs)

    return probs



"""
Methods for analyzing featuresets.

.. autosummary::
   :nosignatures:

   cosine_distance
   cosine_similarity
   distance
   kl_divergence

"""

from math import sqrt, log, acos, pi
from tethne.utilities import nonzero


def kl_divergence(V_a, V_b):
    """
    Calculate Kullback-Leibler distance.

    Uses the smoothing method described in `Bigi 2003
    <http://lvk.cs.msu.su/~bruzz/articles/classification/Using%20Kullback-Leibler%20Distance%20for%20Text%20Categorization.pdf>`_
    to facilitate better comparisons between vectors describing wordcounts.

    Parameters
    ----------
    V_a : list
    V_b : list

    Returns
    -------
    divergence : float
        KL divergence.
    """

    # Find shared features.
    Ndiff = _shared_features(V_a, V_b)

    # aprob and bprob should each sum to 1.0
    aprob = map(lambda v: float(v)/sum(V_a), V_a)
    bprob = map(lambda v: float(v)/sum(V_b), V_b)

    # Smooth according to Bigi 2003.
    aprob, bprob = _smooth(aprob, bprob, Ndiff)

    return sum(map(lambda a, b: (a-b)*log(a/b), aprob, bprob))

def cosine_similarity(F_a, F_b):
    """
    Calculate `cosine similarity
    <http://en.wikipedia.org/wiki/Cosine_similarity>`_ for sparse feature
    vectors.

    Parameters
    ----------
    F_a : :class:`.Feature`
    F_b : :class:`.Feature`

    Returns
    -------
    similarity : float
        Cosine similarity.
    """

    shared = list(F_a.unique & F_b.unique)
    A = [dict(F_a.norm)[i] for i in shared]
    B = [dict(F_b.norm)[i] for i in shared]
    dot = sum(map(lambda a, b: a*b, A, B))
    mag_A = sqrt(sum(map(lambda a: a**2, A)))
    mag_B = sqrt(sum(map(lambda a: a**2, B)))

    return dot / (mag_A + mag_B)

def angular_similarity(F_a, F_b):
    """
    Calculate the `angular similarity
    <http://en.wikipedia.org/wiki/Cosine_similarity#Angular_similarity>`_ for
    sparse feature vectors.

    Unlike `cosine_similarity`, this is a true distance metric.

    Parameters
    ----------
    F_a : :class:`.Feature`
    F_b : :class:`.Feature`

    Returns
    -------
    similarity : float
        Cosine similarity.
    """
    return 1. - (2. *  acos(cosine_similarity(F_a, F_b))) / pi

### Helpers ###


def _shared_features(adense, bdense):
    """
    Number of features in ``adense`` that are also in ``bdense``.
    """
    a_indices = set(nonzero(adense))
    b_indices = set(nonzero(bdense))

    shared = list(a_indices & b_indices)
    diff = list(a_indices - b_indices)
    Ndiff = len(diff)

    return Ndiff

def _smoothing_parameters(aprob, bprob, Ndiff):
    min_a = min(list([list(aprob)[i] for i in nonzero(aprob)]))
    sum_a = sum(aprob)
    min_b = min(list([list(bprob)[i] for i in nonzero(bprob)]))
    sum_b = sum(bprob)

    epsilon = min((min_a/sum_a), (min_b/sum_b)) * 0.001
    gamma = 1 - Ndiff * epsilon

    return gamma, epsilon

def _smooth(aprob, bprob, Ndiff):
    """
    Smooth distributions for KL-divergence according to `Bigi 2003
    <http://link.springer.com/chapter/10.1007%2F3-540-36618-0_22?LI=true>`_.
    """
    gamma, epsilon = _smoothing_parameters(aprob, bprob, Ndiff)

    # Remove zeros.
    in_a = [i for i,v in enumerate(aprob) if abs(v) > 0.0]
    aprob = list([list(aprob)[i] for i in in_a])
    bprob = list([list(bprob)[i]*gamma for i in in_a])

    # Replace zero values with epsilon.
    bprob = list(map(lambda v: v if v != 0. else epsilon, bprob))

    return aprob, bprob
